#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

long myround(double u)
{
  long result = u;
  while (result + 0.5 < u) result += 1;
  while (result - 0.5 > u) result -= 1;
  return result;
}

long long microseconds(void)
{
  struct timeval t;
  gettimeofday(&t,(struct timezone *) 0);
  return t.tv_sec * (long long) 1000000 + t.tv_usec;
}

long long timebase(void) /* only low part on ppc32 */
{
  unsigned long result;
  asm volatile("mftb %0" : "=r"(result));
  return result;
}

double cpufrequency = 0;

double guesstbcycles(void)
{
  long long tb0; long long us0;
  long long tb1; long long us1;

  tb0 = timebase();
  us0 = microseconds();
  do {
    tb1 = timebase();
    us1 = microseconds();
  } while (us1 - us0 < 10000 || tb1 - tb0 < 1000);
  if (tb1 <= tb0) return 0;
  tb1 -= tb0;
  us1 -= us0;
  return (cpufrequency * 0.000001 * (double) us1) / (double) tb1;
}

long doit(char **argv)
{
  char *p;
  int loop;
  double guess1;
  double guess2;
  long tbcycles;

  if (!argv[0]) return 0;
  if (!argv[1]) return 0;
  p = argv[1];
  cpufrequency = strtod(p,&p) * 1000000.0;
  if (cpufrequency <= 0) return 0;

  for (loop = 0;loop < 100;++loop) {
    guess1 = guesstbcycles();
    guess2 = guesstbcycles();
    tbcycles = myround(guess1);
    if (guess1 - tbcycles > 0.1) continue;
    if (tbcycles - guess1 > 0.1) continue;
    if (guess2 - tbcycles > 0.1) continue;
    if (tbcycles - guess2 > 0.1) continue;
    return tbcycles;
  }
  return 0;
}

main(int argc,char **argv)
{
  long tbcycles = doit(argv);
  if (tbcycles > 0) {
    printf("#define PPC_CLOCKS_PER_TICK %ld\n",tbcycles);
    return 0;
  }
  printf("#define PPC_CLOCKS_PER_TICK 16\n");
  return 100;
}
